package org.ak47.android.graphics.ibnrushd.mesh;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import org.ak47.android.graphics.ibnrushd.Constants;
import org.xmlpull.v1.XmlPullParserException;

import android.content.res.XmlResourceParser;
import android.util.Log;

/**
 * Creates meshes on demand
 * @author m0skit0
 */
public class IbnRushdMeshFactory
{
	// Common
	private static final String XML_TAG_COUNT = "count";
	private static final String XML_TAG_ELEMENT = "elem";
	private static final String XML_TAG_ORDER = "order";
	
	// Face
	private static final String XML_TAG_FACE = "face";
	private static final String XML_TAG_F_V0 = "v0";
	private static final String XML_TAG_F_V1 = "v1";
	private static final String XML_TAG_F_V2 = "v2";
	//private static final String XML_TAG_F_N0 = "n0";
	//private static final String XML_TAG_F_N1 = "n1";
	//private static final String XML_TAG_F_N2 = "n2";
	
	// Vertex
	private static final String XML_TAG_VERTEX = "vertex";
	private static final String XML_TAG_V_X = "x";
	private static final String XML_TAG_V_Y = "y";
	private static final String XML_TAG_V_Z = "z";
	private static final String XML_TAG_V_R = "r";
	private static final String XML_TAG_V_G = "g";
	private static final String XML_TAG_V_B = "b";
	private static final String XML_TAG_V_A = "a";
	
	// Current parsing
	private static final int PARSE_NOTHING = -1;
	private static final int PARSE_FACE = 0;
	private static final int PARSE_VERTEX = 1;
	
	private static int parseFaceElementVertices(XmlResourceParser xrp, short[] indexes, int index)
	{
		indexes[++index] = Short.parseShort(xrp.getAttributeValue(null, XML_TAG_F_V0));
		indexes[++index] = Short.parseShort(xrp.getAttributeValue(null, XML_TAG_F_V1));
		indexes[++index] = Short.parseShort(xrp.getAttributeValue(null, XML_TAG_F_V2));
		
		return index;
	}
	
	/*
	private static int parseFaceElementNormals(XmlResourceParser xrp, short[] normals, int index)
	{		
		if (xrp.getAttributeValue(null, XML_TAG_F_N0) != null)
		{
			normals[++index] = Short.parseShort(xrp.getAttributeValue(null, XML_TAG_F_N0));
			normals[++index] = Short.parseShort(xrp.getAttributeValue(null, XML_TAG_F_N1));
			normals[++index] = Short.parseShort(xrp.getAttributeValue(null, XML_TAG_F_N2));
		}
		
		return index;
	}
	*/
	
	private static void parseVertexElement(XmlResourceParser xrp, float[] vertices, float colors[])
	{
		int order = xrp.getAttributeIntValue(null, XML_TAG_ORDER, 0);
		
		int vposition = order * 3;
		vertices[vposition] = xrp.getAttributeFloatValue(null, XML_TAG_V_X, 0);
		vertices[vposition + 1] = xrp.getAttributeFloatValue(null, XML_TAG_V_Y, 0);
		vertices[vposition + 2] = xrp.getAttributeFloatValue(null, XML_TAG_V_Z, 0);

		int cposition = order * 4;
		colors[cposition] = xrp.getAttributeFloatValue(null, XML_TAG_V_R, 0);
		colors[cposition + 1] = xrp.getAttributeFloatValue(null, XML_TAG_V_G, 0);
		colors[cposition + 2] = xrp.getAttributeFloatValue(null, XML_TAG_V_B, 0);
		colors[cposition + 3] = xrp.getAttributeFloatValue(null, XML_TAG_V_A, 0);
	}
	
	/**
	 * Parses an IbnRushd mesh XML
	 * @param xrp XML parser for resource
	 * @return
	 */
	public static IbnRushdMesh parseIbnRushdMesh(XmlResourceParser xrp)
	{
		Log.v(Constants.TAG, "Parsing IbnRushd Mesh XML");
		
		int current_parsing = PARSE_NOTHING;
		
		int num_faces = 0;
		
		short[] indexes = null;
		int indexesIndex = -1;
		
		float[] vertices = null;
		
		float[] colors = null;
		
		// Parse XML
		try
		{
			// While not reached end of document
			while (xrp.getEventType() != XmlResourceParser.END_DOCUMENT)
			{
				String cur_tag = xrp.getName();
				
				// New element
				switch(xrp.getEventType())
				{
					case XmlResourceParser.START_TAG:
						
						// Face group
						if (cur_tag.equals(XML_TAG_FACE))
						{
							num_faces = xrp.getAttributeIntValue(null, XML_TAG_COUNT, 0);
							indexes = new short[num_faces * 3];
							current_parsing = PARSE_FACE;
						}
						
						// Vertex group
						else if (cur_tag.equals(XML_TAG_VERTEX))
						{
							int num_vertices = xrp.getAttributeIntValue(null, XML_TAG_COUNT, 0);
							vertices = new float[num_vertices * 3];
							colors = new float[num_vertices * 4];
							current_parsing = PARSE_VERTEX;
						}
						
						// Element data
						else if (cur_tag.equals(XML_TAG_ELEMENT))
						{
							switch(current_parsing)
							{
								case PARSE_FACE:
									indexesIndex = parseFaceElementVertices(xrp, indexes, indexesIndex);
									//normalsIndex = parseFaceElementVertices(xrp, normals, normalsIndex);
									break;
									
								case PARSE_VERTEX:
									parseVertexElement(xrp, vertices, colors);
									break;
									
								default:
									throw new XmlPullParserException("IbnRushd: parsing unknown group");
							}
						}
						else
						{
							Log.w(Constants.TAG, "Parsing unknown tag: " + cur_tag);
						}
						break;
						
					// Current element closed
					case XmlResourceParser.END_TAG:
						// Closing group
						if (cur_tag.equals(XML_TAG_FACE) || cur_tag.equals(XML_TAG_VERTEX))
						{
							current_parsing = PARSE_NOTHING;
						}
						break;
						
					default:
						break;
				}
				xrp.next();
			}			
		}
		catch (XmlPullParserException e)
		{
			Log.e(Constants.TAG, "Error parsing XML " + xrp.getName(), e);
		}
		catch (IOException e)
		{
			Log.e(Constants.TAG, "I/O error while parsing XML " + xrp.getName(), e);
		}
		finally
		{
			if (xrp != null)
			{
				xrp.close();
			}
		}
		
	    FloatBuffer vertexBuffer = makeFloatBuffer(vertices);
	    FloatBuffer colorBuffer = makeFloatBuffer(colors);
	    ShortBuffer indexBuffer = makeShortBuffer(indexes);
	    FloatBuffer normalBuffer = null; //makeShortBuffer(normals);
	    
	    Log.v(Constants.TAG, "Finished parsing IbnRushd Mesh XML");
		
		return new IbnRushdMesh(vertexBuffer, colorBuffer, normalBuffer, indexBuffer, num_faces);
	}

	// TODO: for debug ONLY
	public static IbnRushdMesh fixedMesh()
	{        
		float vertices[] = //2
        {
			-1.0f, -1.0f, 0,
			1.0f, -1.0f, 0,
			-1.0f, 1.0f, 0,
			1.0f, 1.0f, 0
		};
        
		float maxColor= 1;
		float colors[] = //3
		{
			maxColor,maxColor, 0,maxColor,
			0, maxColor,maxColor,maxColor,
			0, 0, 0,maxColor,
			maxColor, 0,maxColor,maxColor
		};
		
		short indices[] = //4
		{
			0, 3, 1,
			0, 2, 3
		};
		
		int num_faces = 2;
		
		FloatBuffer vertexBuffer = makeFloatBuffer(vertices);
	    FloatBuffer colorBuffer = makeFloatBuffer(colors);
	    ShortBuffer indexBuffer = makeShortBuffer(indices);
	    FloatBuffer normalBuffer = null; //makeShortBuffer(normals);
	    	    
	    Log.v(Constants.TAG, "Finished parsing IbnRushd Mesh XML");
		
		return new IbnRushdMesh(vertexBuffer, colorBuffer, normalBuffer, indexBuffer, num_faces);
	}
	
	private static FloatBuffer makeFloatBuffer(float[] arr)
	{
		FloatBuffer fb = null;
		if (arr != null)
		{
		    ByteBuffer bb = ByteBuffer.allocateDirect(arr.length * 4);
		    bb.order(ByteOrder.nativeOrder());
		    fb = bb.asFloatBuffer();
		    fb.put(arr);
		    fb.position(0);
		}
	    return fb;
	}
	
	private static ShortBuffer makeShortBuffer(short[] arr)
	{
		ShortBuffer sb = null;
		if (arr != null)
		{
			ByteBuffer bb = ByteBuffer.allocateDirect(arr.length * 2);
		    bb.order(ByteOrder.nativeOrder());
		    sb = bb.asShortBuffer();
		    sb.put(arr);
		    sb.position(0);
		}
	    return sb;
	}
}
