#!/usr/bin/perl
# Converts OBJ to IbnRushd XML format
# m0skit0 @ 08/05/2012
# License: GPLv3 (http://www.gnu.org/licenses/gpl-3.0.html)
# Usage: script < obj > xml

# Dependencies:
#	 XML::Simple

# Supports:
#	Vertices
#	Faces (only triangles)
#	Normals
#	Colors for faces

use strict;
use warnings;

use XML::Simple;

# Final XML
my $ref_xml_result;

# Final XML root tag
my $xml_root = 'ibnrushdmesh';

# Parser regular expressions
my $re_is_vertex	= '^v\s+(.*)';		# Vertex
my $re_is_face		= '^f\s+(.*)';		# Face
my $re_is_normal	= '^vn\s+(.*)';		# Normal
my $re_is_mtl_file	= '^mtllib\s+(.*)';	# Material file
my $re_is_mtl		= '^usemtl\s+(.*)';	# Current material

# XML tags:

# Common
my $tag_count = 'count';
my $tag_element = 'elem';
my $tag_order = 'order';

# Vertex
my $tag_vertex			= 'vertex';	# Vertex
my $tag_vertex_x		= 'x';		# Vertex X value
my $tag_vertex_y		= 'y';		# Vertex Y value
my $tag_vertex_z		= 'z';		# Vertex Z value

# Face
my $tag_face	= 'face';				# Face
my @tag_face_v	= ('v0', 'v1', 'v2');	# Vertices
my @tag_face_n	= ('n0', 'n1', 'n2');	# Normals

# Material
my $tag_material	= 'material';			# Material
my @tag_c_rgb		= ('r', 'g', 'b');		# Color components
my $tag_c_alpha		= 'a';					# Alpha

# Normal
my $tag_normal = 'normal';	#Normal

#########
# SUBS
#########

# Vertice
sub process_vertex
{
	my $ref_xml = shift;
	my @v = split(/ /, shift);
	
	if (!$ref_xml->{$tag_vertex})
	{
		$ref_xml->{$tag_vertex} = {};
		$ref_xml->{$tag_vertex}->{$tag_element} = [];
	}
	
	my $ref_cur_vertice =
	{
		$tag_order => $#{$ref_xml->{$tag_vertex}->{$tag_element}} + 1,
		$tag_vertex_x => $v[0],
		$tag_vertex_y => $v[1],
		$tag_vertex_z => $v[2]
	};
	
	push(@{$ref_xml->{$tag_vertex}->{$tag_element}}, $ref_cur_vertice);
	
	return $ref_xml;
}

# Face
sub process_face
{
	my $ref_xml = shift;		# Result XML (hash map reference)
	my @f = split(/ /, shift);	# Vertex indexes for the face
	my $cur_mtl = shift;		# Current material
	
	if (!$ref_xml->{$tag_face})
	{
		$ref_xml->{$tag_face} = {};
		$ref_xml->{$tag_face}->{$tag_element} = [];
	}
	
	$ref_xml->{$tag_face}->{$tag_count}++;
	
	my $ref_cur_face = {};
	
	my $i = 0;
	for(@f)
	{
		my $vertex_index;
		if (/(\d*)\/(\d*)\/(\d*)/)
		{
			$vertex_index = $1 - 1;
			$ref_cur_face->{$tag_face_n[$i]} = $3 - 1;			
		}
		else
		{
			$vertex_index = $_ - 1;
		}
		
		my $ref_vertex = $ref_xml->{$tag_vertex}->{$tag_element}->[$vertex_index];
		
		# Do we have a material to associate?
		if ($cur_mtl)
		{
			my $old_mtl = $ref_vertex->{$tag_material};
			# If vertex already has a material associated
			if ($old_mtl)
			{			
				# If the new material is different than the old one, create a new vertex with the new material
				if ($cur_mtl ne $old_mtl)
				{
					$vertex_index = $#{$ref_xml->{$tag_vertex}->{$tag_element}} + 1;
					my $new_vertex =
					{
						$tag_order => $vertex_index,
						$tag_vertex_x => $ref_vertex->{$tag_vertex_x},
						$tag_vertex_y => $ref_vertex->{$tag_vertex_y},
						$tag_vertex_z => $ref_vertex->{$tag_vertex_z}
					};
					push(@{$ref_xml->{$tag_vertex}->{$tag_element}}, $new_vertex);
					
					# New vertex reference
					$ref_vertex = $new_vertex;
				}
			}
			
			# Associate current material			
			$ref_vertex->{$tag_material} = $cur_mtl;
		}
		
		$ref_cur_face->{$tag_face_v[$i]} = $vertex_index;
		
		$i++;
	};
	
	push(@{$ref_xml->{$tag_face}->{$tag_element}}, $ref_cur_face);
	
	return $ref_xml;
}

# Normal
sub process_normal
{
	my $ref_xml = shift;
	my @v = split(/ /, shift);
	my $order = shift;
	
	if (!$ref_xml->{$tag_normal})
	{
		$ref_xml->{$tag_normal} = {};
		$ref_xml->{$tag_normal}->{$tag_element} = [];
	}
	
	$ref_xml->{$tag_normal}->{$tag_count}++;
	
	my $ref_cur_vertice =
	{
		$tag_order => $order,
		$tag_vertex_x => $v[0],
		$tag_vertex_y => $v[1],
		$tag_vertex_z => $v[2]
	};
	
	push(@{$ref_xml->{$tag_normal}->{$tag_element}}, $ref_cur_vertice);
	
	return $ref_xml;
}

# Material file
sub process_material_file
{
	my $file_name = shift;
	open(FD, '<', $file_name) or die "Couldn't open $file_name\n";
	my @file = <FD>;
	close(FD);
	
	my $cur_mtl;
	my $ref_mtls = {};
	for(@file)
	{
		s/\n//;
		
		# New material
		if (/^newmtl\s+(.*)/)
		{
			$cur_mtl = $1;
		}
		
		# Diffuse
		elsif(/^Kd\s+(.*)/)
		{
			if (!$cur_mtl)
			{
				die "Kd without material name at $_!\n";
			}
			my @kd = split(/ /, $1);
			$ref_mtls->{$cur_mtl} = \@kd;
		}
	}
	
	return $ref_mtls;
}

# Unknown tag
sub unknown_tag
{
	my $arg0 = shift;
	$arg0 =~ /^(\w+)\s/;
	print(STDERR "Tag $1 not supported\n");
}

# Substitutes materials with their properties 
sub process_vertex_materials
{
	my $ref_xml = shift;
	my $ref_mtls = shift;
	
	# Parse all vertex
	for(@{$ref_xml->{$tag_vertex}->{$tag_element}})
	{
		my $cur_vertex = $_;
		my $mtl = $cur_vertex->{$tag_material};
		
		# If vertex has material associated
		if ($mtl)
		{
			my $ref_rgb = $ref_mtls->{$mtl};
			if ($ref_rgb)
			{
				my $i = 0;
				for(@{$ref_rgb})
				{
					$cur_vertex->{$tag_c_rgb[$i++]} = $_;
				}
				$cur_vertex->{$tag_c_alpha} = "0.0";
			}
			else
			{
				die "Could not found definition for material $mtl\n";
			}
			delete($cur_vertex->{$tag_material});
		}
	}
	
	return $ref_xml;
}

#########
# MAIN
#########
	
# Input Wavefront .OBJ file
my @obj = <STDIN>;

# Order
my $cur_vertex = 0;
my $cur_normal = 0;

# Materials hash map
my $ref_mtls = {};

# Current material
my $cur_mtl;

for(@obj)
{
	# Skip comments
	if (!/^#/)
	{
		# Remove newlines
		s/\n//;
		
		# Vertex
		if(/$re_is_vertex/)
		{
			$ref_xml_result = process_vertex($ref_xml_result, $1, $cur_vertex);
			$cur_vertex++;
		}
		
		# Face
		elsif(/$re_is_face/)
		{
			($ref_xml_result) = process_face($ref_xml_result, $1, $cur_mtl);
		}
		
		# Normal
		elsif(/$re_is_normal/)
		{
			$ref_xml_result = process_normal($ref_xml_result, $1, $cur_normal);
			$cur_normal++;
		}
		
		# Material file
		elsif(/$re_is_mtl_file/)
		{
			$ref_mtls = process_material_file($1);
		}
		
		# Current material
		elsif(/$re_is_mtl/)
		{
			$cur_mtl = $1;
		}
		
		# Unknown
		else
		{
			unknown_tag($_);
		}
	}
}

# Update vertices count
$ref_xml_result->{$tag_vertex}->{$tag_count} = $#{$ref_xml_result->{$tag_vertex}->{$tag_element}} + 1;

$ref_xml_result = process_vertex_materials($ref_xml_result, $ref_mtls);

print XMLout($ref_xml_result, RootName => $xml_root);
